/*
 * Copyright (C) 2004 Claudio "KLaN" Ciccani <klan82@cheapnet.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  written by Claudio Ciccani <klan82@cheapnet.it> with the contribute
 *  of Kristof Pelckmans <kristof.pelckmans@antwerpen.be>
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "df_xine.h"



typedef struct
{
     DFXCore *core;
     int      img_area;
     int      width;
     int      height;

} DFXVideoModeData;



static struct
{
     const char            *name;
     DFBSurfacePixelFormat  format;
} pixels[] =
{
     { "UYVY", DSPF_UYVY },
     { "YUY2", DSPF_YUY2 },
     { "I420", DSPF_I420 },
     { "YV12", DSPF_YV12 }
};

static const char *buffermodes[] = 
{
     "single",
     "double",
     NULL,
     "triple",
     NULL
};



static inline int
dfx_log2( int n )
{
     int i = 0;
     
     while (n >> (i + 1))
          i++;     
     return i;
}


static void
dfx_output_cb( void *data, int width, int height,
               double ratio, DFBRectangle* dest_rect )
{
     DFXCore         *this = (DFXCore*) data;
     DFXVideoContext *ctx  = &this->ctx;
     DFXVideoControl *ctl  = &this->ctl;

     if (ctx->caps & DLCAPS_SCREEN_LOCATION)
     {
          /* first time we are called or size changed */
          if (ctx->video.width != width || ctx->video.height != height)
          {
               DBUG( "video size changed:\n" );
               DBUG( "\t-> old: %ix%i\n",
                    ctx->video.width, ctx->video.height);
               DBUG( "\t-> new: %ix%i\n", width, height );

               ctx->config.flags  = (DLCONF_WIDTH | DLCONF_HEIGHT);
               ctx->config.width  = width;
               ctx->config.height = height;

               ctx->layer->SetConfiguration( ctx->layer,
                                   &ctx->config );
               ctx->layer->GetConfiguration( ctx->layer,
                                   &ctx->config);

               ctx->drect.x = 0;
               ctx->drect.y = 0;
               ctx->drect.w = ctx->config.width;
               ctx->drect.h = ctx->config.height;

               ctx->video.width  = width;
               ctx->video.height = height;
          }

          /* first time we are called or ratio changed */
          if (ctx->video.ratio != ratio || ctx->update)
          {
               double  video_ratio;
               float   w, h;

               if (!ctx->update)
               {
                    DBUG( "video ratio changed:\n" );
                    DBUG( "\t-> old %4f\n", (float) ctx->video.ratio );
                    DBUG( "\t-> new %4f\n", (float) ratio );
               }

               video_ratio = ratio;
               if (ctl->ratio && ctl->ratio != ratio)
                    video_ratio = ctl->ratio;

               if (ctx->screen.ratio <= video_ratio)
               {
                   h = (double) ctx->screen.width;
                   /* avoid dividing by 0 */
                   if (video_ratio)
                        h /= video_ratio;

                    ctx->layer->SetScreenLocation( ctx->layer,
                              0.0,
                              ((float) ctx->screen.height - h) /
                               (float) (ctx->screen.height << 1),
                              1.0,
                              h / (float) ctx->screen.height );
               } else
               {
                    w = (double) ctx->screen.height * video_ratio;

                    ctx->layer->SetScreenLocation( ctx->layer,
                              ((float) ctx->screen.width - w) /
                               (float) (ctx->screen.width << 1),
                              0.0,
                              w / (float) ctx->screen.width,
                              1.0 );
               }

               ctx->video.ratio = ratio;
               ctx->update      = 0;
          }

          *dest_rect = ctx->drect;
          
     } else
     {
          /* first time we are called or aspect (size, ratio) changed */
          if (ctx->video.width != width || ctx->video.height != height ||
              ctx->video.ratio != ratio || ctx->update)
          {
               double video_ratio;

               if (!ctx->update)
               {
                    DBUG( "video aspect changed:\n" );
                    DBUG( "\t-> old %ix%i [ratio %4f]\n",
                         ctx->video.width, ctx->video.height,
                         (float) ctx->video.ratio );
                    DBUG( "\t-> new %ix%i [ratio %4f]\n", width,
                         height, (float) ratio );
               }

               ctx->surface->Clear( ctx->surface, 0, 0, 0, 0xff );
               /* Also clear the surfaces held in the back buffers */
               if (ctx->buffermode != DLBM_FRONTONLY)
               {
                    ctx->surface->Flip( ctx->surface, NULL, ctx->flipflags );
                    ctx->surface->Clear( ctx->surface, 0, 0, 0, 0xff );
                    if (ctx->buffermode == DLBM_TRIPLE) 
                    {
                         ctx->surface->Flip( ctx->surface, NULL, ctx->flipflags );
                         ctx->surface->Clear( ctx->surface, 0, 0, 0, 0xff );
                    }
               }
               
               video_ratio = ratio;
               if (ctl->ratio && ctl->ratio != ratio)
                    video_ratio = ctl->ratio;

               if (ctx->screen.ratio <= video_ratio)
               {
                    ctx->drect.w = ctx->screen.width;
                    ctx->drect.h = (double) ctx->screen.width / video_ratio;
               } else
               {
                    ctx->drect.w = (double) ctx->screen.height * video_ratio;
                    ctx->drect.h = ctx->screen.height;
               }

               ctx->drect.x = (ctx->screen.width  - ctx->drect.w) >> 1;
               ctx->drect.y = (ctx->screen.height - ctx->drect.h) >> 1;

               ctx->video.width  = width;
               ctx->video.height = height;
               ctx->video.ratio  = ratio;
               ctx->update       = 0;
          }

          if (ctx->buffer) /* using hwstretchblit */
          {
               ctx->buffer->GetSize( ctx->buffer,
                         &dest_rect->w, &dest_rect->h );
               dest_rect->x = 0;
               dest_rect->y = 0;
          } else
               *dest_rect = ctx->drect;
     }
}


static int
dfx_frame_cb( void *data )
{
     DFXVideoContext *ctx = &((DFXCore*) data)->ctx;
     int              dw;
     int              dh;
     int              sw;
     int              sh;

     if (ctx->buffer)
     {
          ctx->surface->GetSize( ctx->surface, &dw, &dh );
          ctx->buffer->GetSize( ctx->buffer, &sw, &sh );

          if (dw != sw || dh != sh)
               ctx->surface->StretchBlit( ctx->surface, 
                         ctx->buffer, NULL, &ctx->drect );
          else
               ctx->surface->Blit( ctx->surface, ctx->buffer,
                         NULL, ctx->drect.x, ctx->drect.y );
     }

     ctx->surface->Flip( ctx->surface, NULL, ctx->flipflags );

     return 0;
}


static void
dfx_test_layer( DFXCore *this, IDirectFBDisplayLayer *layer,
                DFBDisplayLayerConfig *config )
{
     DFXVideoContext       *ctx = &this->ctx;
     DFBDisplayLayerConfig  cfg;
     DFBSurfacePixelFormat  fmt  = 0;
     DFBResult              err;
     int                    i;

     config->flags = 0;
     layer->GetConfiguration( layer, &cfg );

     SAY( "-> checking if %s-buffering is supported...",
               buffermodes[dfx_log2( ctx->buffermode )] );
     
     /* buffermpde test */
     cfg.flags      = DLCONF_BUFFERMODE;
     cfg.buffermode = ctx->buffermode;

     err = layer->TestConfiguration( layer, &cfg, NULL );
     if (err == DFB_OK)
     {
          config->flags      |= DLCONF_BUFFERMODE;
          config->buffermode  = ctx->buffermode;
          if (this->verbosity)
               fprintf( stderr, "yes\n" );
     } else
     {
          if (this->verbosity)
               fprintf( stderr, "no\n" );
     }

     /* pixelformat test */
     cfg.flags = DLCONF_PIXELFORMAT;

     for (i = 0; i < (sizeof( pixels) / sizeof( pixels[0] )); i++)
     {
          SAY( "-> checking if %s is supported...", pixels[i].name );
          cfg.pixelformat = pixels[i].format;
          err = layer->TestConfiguration( layer, &cfg, NULL );
          if (err == DFB_OK)
          {
               fmt = cfg.pixelformat;
               if (this->verbosity)
                    fprintf( stderr, "yes\n" );
          } else
          {
               if (this->verbosity)
                    fprintf( stderr, "no\n" );
          }
     }

     if (fmt)
     {
          config->flags       |= DLCONF_PIXELFORMAT;
          config->pixelformat  = fmt; /* use last working format */
     }
}



static DFBEnumerationResult
dfx_scan_layers( unsigned id, DFBDisplayLayerDescription desc, void *data )
{
     DFXCore               *this   = (DFXCore*) data;
     DFXVideoContext       *ctx    = &this->ctx;
     IDirectFBDisplayLayer *cur    = NULL;
     DFBDisplayLayerConfig  config;
     DFBResult              err;

     if (id == DLID_PRIMARY)
          return DFENUM_OK;
     
     SAY( "trying layer %i\n", id );

     if (!(desc.caps & DLCAPS_SURFACE))
     {
          SAY( "-> not usable!\n" );
          return DFENUM_OK;
     }

     SAY( "-> has a surface\n" );

     if (desc.caps & DLCAPS_SCREEN_LOCATION)
          SAY( "-> supports screen location\n" );

     SAY( "-> trying to access..." );

     err = this->dfb->GetDisplayLayer( this->dfb, id, &cur );

     if (err == DFB_OK)
          err = cur->SetCooperativeLevel( cur, DLSCL_EXCLUSIVE );

     if (err != DFB_OK)
     {
          if (this->verbosity)
               fprintf( stderr, "failed\n" );
          goto failure;
     }

     if (this->verbosity)
          fprintf( stderr, "ok\n" );

     dfx_test_layer( this, cur, &config );

     if (!(config.flags & DLCONF_PIXELFORMAT))
     {
          /* no usable pixelformat found */
          SAY( "-> not enough!\n" );
          goto failure;
     }

     SAY( "using layer %i [%s]\n", id, desc.name );
     this->ctx.lid      = id;
     this->ctx.layer    = cur;
     this->ctx.caps     = desc.caps;
     
     /* set configuration */
     cur->SetConfiguration( cur, &config );
     cur->GetConfiguration( cur, &config );

     if (config.buffermode != DLBM_FRONTONLY)
          ctx->flipflags |= DSFLIP_ONSYNC; /* Matrox BES needs this */

     return DFENUM_CANCEL;

failure:
     if (cur)
          cur->Release( cur );

     return DFENUM_OK;
}


static void
dfx_open_video( DFXCore *this )
{
     DFXVideoContext      *ctx    = &this->ctx;
     dfb_visual_t          visual;
     dfb_frame_callback_t  fcb;

     this->bg->GetSize( this->bg, &ctx->screen.width, &ctx->screen.height );
     ctx->screen.ratio = (double) ctx->screen.width /
                         (double) ctx->screen.height;
     
     if (this->ctx.lid >= 0)
     {
               DFBDisplayLayerDescription desc;
               DFBDisplayLayerConfig      config;

               SAY( "forced to use layer %i\n", this->ctx.lid );

               DFBCHECK(this->dfb->GetDisplayLayer( this->dfb,
                              this->ctx.lid, &ctx->layer));
               DFBCHECK(ctx->layer->SetCooperativeLevel( ctx->layer,
                                        DLSCL_EXCLUSIVE ));

               dfx_test_layer( this, ctx->layer, &config );
               ctx->layer->SetConfiguration( ctx->layer, &config );
               ctx->layer->GetConfiguration( ctx->layer, &config );

               if (config.buffermode != DLBM_FRONTONLY)
                    ctx->flipflags |= DSFLIP_ONSYNC; /* Matrox BES needs this */

               ctx->layer->GetDescription( ctx->layer, &desc );
               ctx->caps = desc.caps;

               if (ctx->caps & DLCAPS_SCREEN_LOCATION)
                    ctx->layer->SetScreenLocation( ctx->layer,
                                             0.0, 0.0, 1.0, 1.0 );
               else
                    ctx->config = config;

               DFBCHECK(ctx->layer->GetSurface( ctx->layer,
                                   &ctx->surface ));
               ctx->surface->Clear( ctx->surface, 0, 0, 0, 0xff );
               ctx->surface->Flip( ctx->surface, NULL, DSFLIP_WAITFORSYNC );
     } else
     {
          IDirectFBScreen *screen;
          
          SAY( "scanning layers for a suitable one\n" );

          DFBCHECK(this->primary->GetScreen( this->primary, &screen ));
          DFBCHECK(screen->EnumDisplayLayers( screen,
                         dfx_scan_layers, (void*) this ));

          if (ctx->lid < 0) {
               DFBDisplayLayerConfig config;
               
               SAY( "no suitable layer found\n" );
               SAY( "falling back to primary layer :(\n" );

               this->primary->SetCooperativeLevel( this->primary, 
                                                   DLSCL_EXCLUSIVE );
               
               config.flags      = DLCONF_BUFFERMODE;
               config.buffermode = ctx->buffermode;
               this->primary->SetConfiguration( this->primary, &config );
               this->primary->GetConfiguration( this->primary, &config );

               /*if (config.buffermode != DLBM_FRONTONLY)
                    ctx->flipflags |= DSFLIP_ONSYNC;*/

               this->primary->AddRef( this->primary );
               ctx->layer = this->primary;
               ctx->lid   = 0;
               ctx->caps  = 0;
          } 
          
          if (ctx->caps & DLCAPS_SCREEN_LOCATION)
                ctx->layer->SetScreenLocation( ctx->layer,
                                               0.0, 0.0, 1.0, 1.0 );
          else
                ctx->layer->GetConfiguration( ctx->layer,
                                              &ctx->config );

          DFBCHECK(ctx->layer->GetSurface( ctx->layer,
                                           &ctx->surface ));
          ctx->surface->Clear( ctx->surface, 0, 0, 0, 0xff );
          ctx->surface->Flip( ctx->surface, NULL, DSFLIP_WAITFORSYNC );
          
          screen->Release( screen );
     }

     visual.surface   = ctx->surface;
     visual.output_cb = dfx_output_cb;
     visual.cdata     = (void*) this;

     this->vo = xine_open_video_driver( this->xine, "DFB",
                    XINE_VISUAL_TYPE_DFB, (void*) &visual );
     if (!this->vo)
          FATAL( "couldn't open video driver 'DFB'; have you installed it?\n" );

     fcb.frame_cb = dfx_frame_cb;
     fcb.cdata    = (void*) this;

     xine_port_send_gui_data( this->vo,
                              XINE_GUI_SEND_TRANSLATE_GUI_TO_VIDEO,
                              (void*) &fcb );
}


static void
dfx_open_audio( DFXCore *this )
{
     const char* const *ao_list;
     const char        *ao_driver;

     ao_list = xine_list_audio_output_plugins( this->xine );

     ao_driver = xine_config_register_string( this->xine, "audio.driver",
                         ao_list[0], "Audio driver to use",
                         NULL, 0, NULL, NULL );

     this->ao = xine_open_audio_driver( this->xine, ao_driver, NULL );
     
     if (!this->ao)
          FATAL( "couldn't open audio driver '%s'\n", ao_driver );
}


void
dfx_context_create( DFXCore *this )
{
     this->xine = xine_new();
     if (!this->xine)
          FATAL( "failed xine initialization (in xine_new())\n" );

     if (!getenv( "XINERC" ))
     {
          char *xined;
          xined = DX_TOSTRING( "%s/.xine", xine_get_homedir() );
          mkdir( xined, 755 );
          this->cfg = DX_TOSTRING( "%s/config", xined );
          free(xined);
     } else
          this->cfg = DX_STRDUP( getenv( "XINERC" ));

     if (this->cfg)
          xine_config_load( this->xine, this->cfg );

     xine_init( this->xine );
     xine_engine_set_param( this->xine, XINE_ENGINE_PARAM_VERBOSITY,
                    this->verbosity );

     dfx_open_video( this );
     dfx_open_audio( this );

     this->stream = xine_stream_new( this->xine, this->ao, this->vo );
     if (!this->stream)
          FATAL( "failed stream initialization (in xine_stream_new())\n" );

     xine_set_param( this->stream, XINE_PARAM_VERBOSITY, this->verbosity );
     xine_set_param( this->stream, XINE_PARAM_AUDIO_MUTE, 0 );
     xine_set_param( this->stream, XINE_PARAM_AUDIO_CHANNEL_LOGICAL, -1 );
}


static DFBEnumerationResult
dfx_mode_cb( int width, int height, int bpp, void *data )
{
     DFXVideoModeData *dta  = (DFXVideoModeData*) data;
     DFXCore          *this = dta->core;

     SAY( "probing video mode %ix%i@%i...",
               width, height, bpp );

     if (!dta->width && !dta->height)
     {
          dta->width  = width;
          dta->height = height;
          if (this->verbosity)
               fprintf( stderr, "accepted\n" );
     } else
     if (abs( dta->img_area - (width * height) ) <
         abs( dta->img_area - (dta->width * dta->height) ))
     {
          dta->width  = width;
          dta->height = height;
          if (this->verbosity)
               fprintf( stderr, "accepted\n" );
     } else
     {
          if (this->verbosity)
               fprintf( stderr, "not better\n" );
     }

     return DFENUM_OK;
}


void
dfx_context_update( DFXCore *this, int width, int height )
{
     DFXVideoContext       *ctx  = &this->ctx;
     DFBSurfacePixelFormat  fmt;
     DFBAccelerationMask    mask;
     DFBResult              err;
     DFXVideoModeData       data =
     {
          .core     = this,
          .img_area = width * height,
          .width    = 0,
          .height   = 0
     };

     if (ctx->lid != DLID_PRIMARY &&
         (ctx->caps & DLCAPS_SCREEN_LOCATION))
          return;

     if (ctx->screen.width  == width &&
         ctx->screen.height == height)
          return;

     this->bg->GetAccelerationMask( this->bg, this->bg, &mask );
     this->bg->GetPixelFormat( this->bg, &fmt );

     if (mask & DFXL_STRETCHBLIT)
     {
          DFBSurfaceDescription dsc;

          ONCE( "hwstretchblit detected and enabled\n" );
          
          if (ctx->buffer)
               ctx->buffer->Release( ctx->buffer );

          dsc.flags       = (DSDESC_WIDTH | DSDESC_HEIGHT |
                       DSDESC_CAPS | DSDESC_PIXELFORMAT);
          dsc.caps        = DSCAPS_VIDEOONLY;
          dsc.width       = width;
          dsc.height      = height;
          dsc.pixelformat = fmt;

          SAY( "setting up temporary surface for hwstretchblit..." );

          err = this->dfb->CreateSurface( this->dfb, &dsc, &ctx->buffer );
          if (err != DFB_OK)
          {
               if (this->verbosity)
                    fprintf( stderr, "failed\n" );
               SAY( "hwstretchblit disabled\n" );
               xine_port_send_gui_data( this->vo,
                         XINE_GUI_SEND_DRAWABLE_CHANGED,
                         (void*) ctx->surface );
               ctx->buffer = NULL;
          } else
          {
               if (this->verbosity)
                    fprintf( stderr, "ok\n" );
               xine_port_send_gui_data( this->vo,
                         XINE_GUI_SEND_DRAWABLE_CHANGED,
                         (void*) ctx->buffer );
          }
     
     } else /* no hwstretchblit */
     {     
          this->dfb->EnumVideoModes( this->dfb,
                    dfx_mode_cb, (void*) &data );

          if (ctx->screen.width  != data.width ||
                  ctx->screen.height != data.height)
          {
               SAY( "setting video mode to %ix%i@%i...",
                    data.width, data.height, DFB_BITS_PER_PIXEL( fmt ));
          
               err = this->dfb->SetVideoMode( this->dfb, data.width,
                         data.height, DFB_BITS_PER_PIXEL( fmt ));
               this->dfb->WaitForSync( this->dfb );

               if (this->verbosity)
                    fprintf( stderr, (err != DFB_OK) ? "failed\n" : "ok\n" );

               this->bg->GetSize( this->bg, &ctx->screen.width,
                            &ctx->screen.height );
               ctx->screen.ratio = (double) ctx->screen.width /
                              (double) ctx->screen.height;
          
               ctx->update = 1;
          } else
               SAY( "video mode %ix%i@%i already set\n",
                    data.width, data.height, DFB_BITS_PER_PIXEL( fmt ) );
     }
}


void
dfx_context_set_gmode( DFXCore *this, int mode )
{
     DFXVideoContext *ctx  = &this->ctx;
     xine_post_out_t *audio_source;
     int              has_video;

     if (ctx->gmode == mode)
          return;
     
     if (!this->post_plugin)
     {
          const char* const *post_list;

          post_list = xine_list_post_plugins_typed( this->xine,
                         XINE_POST_TYPE_AUDIO_VISUALIZATION );
          this->post_plugin = xine_config_register_string( this->xine,
                         "gui.post_audio_plugin", post_list[0],
                         "Audio visualization plugin",
                         NULL, 0, NULL, NULL );
     }

     audio_source = xine_get_audio_source( this->stream );
     has_video    = xine_get_stream_info( this->stream,
                         XINE_STREAM_INFO_HAS_VIDEO );

     if (mode == DGM_VIDEO)
     {
          xine_post_wire_audio_port( audio_source, this->ao );

          if (this->post)
          {
               xine_post_dispose( this->xine, this->post );
               this->post = NULL;
          }

          if (has_video)
               xine_set_param( this->stream,
                         XINE_PARAM_IGNORE_VIDEO, 0 );

          DBUG( "graphics mode set to Video\n" );
     } else
     if (mode == DGM_ANIM)
     {
          if (!this->post)
          {
               this->post = xine_post_init( this->xine, this->post_plugin,
                                        0, &this->ao, &this->vo );
               if (!this->post)
               {
                    SAY( "initialization failed for post plugin'%s'\n",
                         this->post_plugin );
                    return;
               }
          }

          if (has_video)
               xine_set_param( this->stream,
                         XINE_PARAM_IGNORE_VIDEO, 1 );

          xine_post_wire_audio_port( audio_source,
                         this->post->audio_input[0] );

          DBUG( "graphics mode set to Audio Visualization\n" );
     } else
     {
          DBUG( "unknown graphics mode %i\n", mode );
          return;
     }

     ctx->gmode = mode;
}


void
dfx_context_release( DFXCore *this )
{
     DFXVideoContext *ctx = &this->ctx;
     
     if (this->stream)
          xine_dispose( this->stream );

     if (this->post)
          xine_post_dispose( this->xine, this->post );

     if (this->vo)
          xine_close_video_driver( this->xine, this->vo );

     if (this->ao)
          xine_close_audio_driver( this->xine, this->ao );

     if (this->xine)
     {
          if (this->cfg)
          {
               xine_config_save( this->xine, this->cfg );
               free( this->cfg );
          }
          
          xine_exit( this->xine );
     }

     if (ctx->buffer)
          ctx->buffer->Release( ctx->buffer );
     
     if (ctx->surface)
          ctx->surface->Release( ctx->surface );

     if (ctx->layer)
          ctx->layer->Release( ctx->layer );
}

